/* Code to get discid for a cddb query.

  *** Should work in Linux, Solaris, FreeBSD, and OpenBSD ***

  Copyright (c) 1998-2000 Jeremy D. Zawodny <Jeremy@Zawodny.com>

  This software is covered by the GPL.

  Is is based on code found in:

    To: code-review at azure dot humbug dot org dot au 
    Subject: CDDB database reader 
    From: Byron Ellacott <rodent at route-qn dot uqnga dot org dot au> 
    Date: Fri, 5 Jun 1998 17:32:40 +1000 

  Portability code derived from python-cddb (http://cddb-py.sourceforge.net/)

  $Id: discid.c,v 1.3 2004/06/05 17:08:38 muth Exp $

*/

/* Stripped net code, 'cause I only care about the discid */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>


#define CDDB_DEFAULT_CDROM_FLAGS (O_RDONLY | O_NONBLOCK)

#if defined(__linux__)
#include <linux/cdrom.h>

#elif defined(sun)
#include <sys/cdio.h>

#elif defined(__FreeBSD__)
#include <sys/cdio.h>

#elif defined(__OpenBSD__)
#include <sys/cdio.h>

#else
#error "unknown platform"
#endif

/* 
 * Since FreeBSD has identical support but different names for lots
 * of these structs and constants, we'll just #define CDDB_WHATEVER
 * so that we don't have to repeat the code.
 */

#ifdef __FreeBSD__

#define CDDB_TOC_HEADER_STRUCT ioc_toc_header 
#define CDDB_STARTING_TRACK_FIELD starting_track 
#define CDDB_ENDING_TRACK_FIELD ending_track
#define CDDB_READ_TOC_HEADER_FLAG CDIOREADTOCHEADER 
#define CDDB_TOC_ENTRY_STRUCT ioc_read_toc_single_entry 
#define CDDB_TRACK_FIELD track 
#define CDDB_FORMAT_FIELD address_format 
#define CDDB_MSF_FORMAT CD_MSF_FORMAT 
#define CDDB_ADDR_FIELD entry.addr 
#define CDDB_READ_TOC_ENTRY_FLAG CDIOREADTOCENTRY 
#define CDDB_CDROM_LEADOUT 0xaa 

#elif defined(__OpenBSD__)

#define CDDB_TOC_HEADER_STRUCT ioc_toc_header
#define CDDB_STARTING_TRACK_FIELD starting_track
#define CDDB_ENDING_TRACK_FIELD ending_track
#define CDDB_READ_TOC_HEADER_FLAG CDIOREADTOCHEADER
#define CDDB_TOC_ENTRY_STRUCT ioc_read_toc_entry
#define CDDB_TRACK_FIELD starting_track
#define CDDB_FORMAT_FIELD address_format
#define CDDB_MSF_FORMAT CD_MSF_FORMAT
#define CDDB_ADDR_FIELD data->addr
#define CDDB_READ_TOC_ENTRY_FLAG CDIOREADTOCENTRIES
#define CDDB_CDROM_LEADOUT 0xaa

#else /* Linux and Solaris */

#define CDDB_TOC_HEADER_STRUCT cdrom_tochdr 
#define CDDB_STARTING_TRACK_FIELD cdth_trk0 
#define CDDB_ENDING_TRACK_FIELD cdth_trk1 
#define CDDB_READ_TOC_HEADER_FLAG CDROMREADTOCHDR 
#define CDDB_TOC_ENTRY_STRUCT cdrom_tocentry
#define CDDB_TRACK_FIELD cdte_track 
#define CDDB_FORMAT_FIELD cdte_format 
#define CDDB_MSF_FORMAT CDROM_MSF 
#define CDDB_ADDR_FIELD cdte_addr 
#define CDDB_READ_TOC_ENTRY_FLAG CDROMREADTOCENTRY 
#define CDDB_CDROM_LEADOUT CDROM_LEADOUT

#endif 


struct toc
{
    int min, sec, frame;
} cdtoc[100];

int read_toc(char *cdrom_name)
{
    int drive = open(cdrom_name, CDDB_DEFAULT_CDROM_FLAGS);
    struct CDDB_TOC_HEADER_STRUCT tochdr;
    struct CDDB_TOC_ENTRY_STRUCT tocentry;
    int i;
#if defined(__OpenBSD__)
    struct cd_toc_entry data;
#endif

    ioctl(drive, CDDB_READ_TOC_HEADER_FLAG, &tochdr);

    for (i = tochdr.CDDB_STARTING_TRACK_FIELD; i <= tochdr.CDDB_ENDING_TRACK_FIELD; i++)
    {
        tocentry.CDDB_TRACK_FIELD = i;
        tocentry.CDDB_FORMAT_FIELD = CDDB_MSF_FORMAT;

#if defined(__OpenBSD__)
        tocentry.data = &data;
        tocentry.data_len = sizeof(data);
#endif

        ioctl(drive, CDDB_READ_TOC_ENTRY_FLAG, &tocentry);
        cdtoc[i-1].min = tocentry.CDDB_ADDR_FIELD.msf.minute;
        cdtoc[i-1].sec = tocentry.CDDB_ADDR_FIELD.msf.second;
        cdtoc[i-1].frame = tocentry.CDDB_ADDR_FIELD.msf.frame;
        cdtoc[i-1].frame += cdtoc[i-1].min*60*75;
        cdtoc[i-1].frame += cdtoc[i-1].sec*75;
    }

    tocentry.CDDB_TRACK_FIELD = 0xAA;
    tocentry.CDDB_FORMAT_FIELD = CDDB_MSF_FORMAT;
    ioctl(drive, CDDB_READ_TOC_ENTRY_FLAG, &tocentry);
    cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].min = tocentry.CDDB_ADDR_FIELD.msf.minute;
    cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].sec = tocentry.CDDB_ADDR_FIELD.msf.second;
    cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].frame = tocentry.CDDB_ADDR_FIELD.msf.frame;
    cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].frame += cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].min*60*75;
    cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].frame += cdtoc[tochdr.CDDB_ENDING_TRACK_FIELD].sec*75;
    close(drive);
    return tochdr.CDDB_ENDING_TRACK_FIELD;
}

unsigned int cddb_sum(int n)
{
    unsigned int ret;

    ret = 0;
    while (n > 0) {
      ret += (n % 10);
      n /= 10;
    }
    return ret;
}

unsigned int cddb_discid(int tot_trks)
{
    unsigned int i, t = 0, n = 0;

    i = 0;
    while (i < tot_trks)
    {
        n = n + cddb_sum((cdtoc[i].min * 60) + cdtoc[i].sec);
        i++;
    }

    t = ((cdtoc[tot_trks].min * 60) + cdtoc[tot_trks].sec) - ((cdtoc[0].min * 60) + cdtoc[0].sec);

    return ((n % 0xff) << 24 | t << 8 | tot_trks);
}

int main(int argc, char *argv[])
{
    unsigned int discid;
    int tracks, i;

    if( argc !=  2 )
    {
        fprintf(stderr,
                "usage: \n"
                "discid cdrom-device\n"
                "read toc\n");
        return -1;
    }

    tracks = read_toc(argv[1]);
    discid = cddb_discid(tracks);
    printf("%08x %d", discid, tracks);

    for (i = 0; i < tracks; i++)
    {
        printf(" %d", cdtoc[i].frame);
    }
    
    printf(" %d\n", (cdtoc[tracks].frame-cdtoc[0].frame)/75);
    printf("\n");
    return 0;
}
